#include <libxode.h>
#include <jabber/jabber.h>
#include <etherx.h>
#include <faim/aim.h>

extern ehandler ehandler_at;
extern char *at_name;
extern ppdb PPDB;

/* auth.c */
typedef struct at_login_struct {
    char *user;
    char *pass;
    jid from;
    jid to;
    jpacket jp;
} _at_login, *at_login;

typedef struct at_buddy_struct {
    jid from;
    jid to;
	time_t t;
	struct at_buddy_struct *next;
} _at_buddy, *at_buddy;

/* sessions.c */
typedef struct {
    pth_message_t head; /* the standard pth message header */
    jpacket p;
} _jpq, *jpq;

typedef struct sessions_struct
{
	jid j;
	jid from;
	struct aim_session_t ass;
    pool p;
	pth_msgport_t mp_to;
	int exit_flag;
	int loggedin;
	char *buddylist;
	char *profile;
	at_buddy buddies;
	struct sessions_struct *next;
} _session, *session;

void at_sleeper_main(void *arg);

at_buddy at_auth_find_buddy(at_buddy head, jid j);
void at_auth_subscribe_buddy(jpacket jp);
void at_buddy_find_stale(size_t t);
void at_auth_add_buddy(jpacket jp);
int at_auth_user(xmlnode xp);
void at_auth_register(jpacket jp);
void at_subscribe_auth(jid to, jid from, char *user, char *pass);
char *at_auth_get_buddies(session s, jid from);


/* external.c */
void at_init();
void at_etherx_receive(epacket e);


session at_session_create(char *user, char *pword, jid j, jid from);
int at_session_isdone(void *arg);
void at_session_end(session s);
session at_session_find_by_jid(jid j);
void at_session_deliver(session s, xmlnode x, jid from);
void at_aim_session_deliver(session s, xmlnode x, jid from);
void at_session_unlink(session s);
void at_session_buddy_unlink(session s, char *user);
int at_aim_session_parser(session s, jpacket jp);

/* control.c */
int at_trans_stop(void);

/* parser.c */
void at_parse_packet(xmlnode x);
int at_parse_message(jpacket jp);

/* utils.c */
void at_psend(pth_msgport_t mp, jpacket p);
char *str_to_UTF8(pool p,unsigned char *in);
unsigned char *UTF8_to_str(pool p,unsigned char *in);


/* register.c */
void at_auth_register(jpacket jp);

/* aimserver.c */
int at_parse_misses(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_serverready(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_handleredirect(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_auth_error(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_auth_success(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_parse_userinfo(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_authsvrready(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_pwdchngdone(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_parse_oncoming(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_parse_offgoing(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_parse_incoming_im(struct aim_session_t *ass, struct command_rx_struct *command, ...);
int at_chatnav_info(struct aim_session_t *sess, struct command_rx_struct *command, ...);
int at_chat_join(struct aim_session_t *sess, struct command_rx_struct *command, ...);
int at_chat_leave(struct aim_session_t *sess, struct command_rx_struct *command, ...);
int at_chat_infoupdate(struct aim_session_t *sess, struct command_rx_struct *command, ...);
int at_chat_incomingmsg(struct aim_session_t *sess, struct command_rx_struct *command, ...);
int at_parse_motd(struct aim_session_t *sess, struct command_rx_struct *command, ...);
int at_parse_authresp(struct aim_session_t *sess, struct command_rx_struct *command, ...);

struct aim_conn_t *_aim_select(struct  aim_session_t *ass, pth_event_t ev, int *status);
char *at_normalize(char *s);

/* unknown.c */
void *at_unknown_main(void *arg);


/* NEW */
int at_parse_connerr(struct aim_session_t *sess, struct command_rx_struct *command, ...);
