#include <aimtrans.h>

char *at_normalize(char *s)
{
    char *new, *old;

    if(s == NULL)
        return;

    log_debug(ZONE,"normalizing %s",s);

    new = old = s;
    while(*old)
    {
        if(*old == ' ')
        {
            old++;
            continue;
        }
        *new++ = *old++;
    }
    *new = '\0';

    log_debug(ZONE,"normalized %s",s);
    return s;
}

int at_serverready(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
  switch (command->conn->type)
    {
    case AIM_CONN_TYPE_BOS:
      aim_bos_reqrate(ass, command->conn); /* request rate info */
      aim_bos_ackrateresp(ass, command->conn);  /* ack rate info response -- can we say timing? */
      aim_bos_setprivacyflags(ass, command->conn, 0x00000003);
      
#if 0
      aim_bos_reqpersonalinfo(command->conn);
#endif
      
      aim_bos_reqservice(ass, command->conn, AIM_CONN_TYPE_ADS); /* 0x05 == Advertisments */
	  aim_setversions(ass, command->conn);

#if 0
      aim_bos_reqrights(NULL);
      aim_bos_reqbuddyrights(NULL);
      aim_bos_reqlocaterights(NULL);
      aim_bos_reqicbmparaminfo(NULL);
#endif
      
      /* set group permissions */
      aim_bos_setgroupperm(ass, command->conn, 0x1f);
      break;
    case AIM_CONN_TYPE_CHATNAV:
      //fprintf(stderr, "ac: chatnav: got server ready\n");
	  aim_conn_addhandler(ass, command->conn, AIM_CB_FAM_CTN, AIM_CB_CTN_INFO, at_chatnav_info, 0);
      aim_bos_reqrate(ass, command->conn);
      aim_bos_ackrateresp(ass, command->conn);
      aim_chatnav_clientready(ass, command->conn);
      aim_chatnav_reqrights(ass, command->conn);
      break;
	case AIM_CONN_TYPE_CHAT:
      aim_conn_addhandler(ass, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERJOIN, at_chat_join, 0);
      aim_conn_addhandler(ass, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_USERLEAVE, at_chat_leave, 0);
      aim_conn_addhandler(ass, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_ROOMINFOUPDATE, at_chat_infoupdate, 0);
      aim_conn_addhandler(ass, command->conn, AIM_CB_FAM_CHT, AIM_CB_CHT_INCOMINGMSG, at_chat_incomingmsg, 0);
      aim_bos_reqrate(ass, command->conn);
      aim_bos_ackrateresp(ass, command->conn);
      aim_chat_clientready(ass, command->conn);
      break;
    default:
      log_warn("[AIM]", "at: unknown connection type on Server Ready\n");
    }
  return 1;
}

/*
  handleredirect()...

  This, of course, handles Service Redirects from OSCAR.

  Should get passed in the following:
     struct command_rx_struct *command
       the raw command data
     int serviceid
       the destination service ID
     char *serverip
       the IP address of the service's server
     char *cookie
       the raw auth cookie
 */
int at_handleredirect(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
  va_list ap;
  int serviceid;
  char *ip;
  char *cookie;
  session s;
  xmlnode x;

  /* this is the new buddy list */
  char buddies[] = "";
  /* this is the new profile */
  char profile[] = "";  

  va_start(ap, command);
  serviceid = va_arg(ap, int);
  ip = va_arg(ap, char *);
  cookie = va_arg(ap, char *);

  switch(serviceid)
    {
    case 0x0005: /* Advertisements */
      /*
       * The craziest explanation yet as to why we finish logging in when
       * we get the advertisements redirect, of which we don't use anyway....
       *                    IT WAS EASY!
       */

	  s = ass->aux_data;
      /* send the buddy list and profile (required, even if empty) */
	  s->buddylist = at_auth_get_buddies(s, s->j);
	  if(s->buddylist==NULL)
	{
		log_debug(ZONE, "[AIM] We have a NULL budylist!\n");
	    aim_bos_setbuddylist(ass, command->conn, buddies);
	}
	  else
	{
		log_debug(ZONE, "[AIM] using generated buddylist %s\n", s->buddylist);
		aim_bos_setbuddylist(ass, command->conn, s->buddylist);
	}
	  aim_bos_setprofile(ass, command->conn, profile, NULL, AIM_CAPS_CHAT);
      /*I want to do this eventually
	   * aim_bos_setprofile(ass, command->conn, s->profile, NULL);
	   */

      /* send final login command (required) */
      aim_bos_clientready(ass, command->conn); /* tell BOS we're ready to go live */
      /* Set our latency... */
      aim_conn_setlatency(command->conn, 1);
	  log_debug(ZONE, "[AIM] Logged in\n");

	  /* This is handled above and previosly now
	   * 
	   * at_auth_get_buddies(&s->ass, s->j);
	   */
	  s->loggedin = 1;
      x = jutil_presnew(JPACKET__AVAILABLE, jid_full(s->j), "Connected");
      xmlnode_put_attrib(x, "from", jid_full(s->from));
      ehandler_send(ehandler_at, x, s->j->server);

      /* you should now be ready to go */
      //log_debug(ZONE, "[AIM] You are now officially online. (%s)\n", ip);      

      break;
    case 0x0007: /* Authorizer */
      {
	struct aim_conn_t *tstconn;
	/* Open a connection to the Auth */
	tstconn = aim_newconn(ass, AIM_CONN_TYPE_AUTH, ip);
	if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR) )
	  log_warn("[AIM]", "ac: unable to reconnect with authorizer\n");
	else
	  /* Send the cookie to the Auth */
	  aim_auth_sendcookie(ass, tstconn, cookie);

      }  
      break;
    case 0x000d: /* ChatNav */
      {
	struct aim_conn_t *tstconn = NULL;
	tstconn = aim_newconn(ass, AIM_CONN_TYPE_CHATNAV, ip);
	if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
	  {
	    //fprintf(stderr, "ac: unable to connect to chatnav server\n");
        log_debug(ZONE, "[AIM] Could not connect to chatnav server");
        if (tstconn) aim_conn_kill(ass, &tstconn);
	    return 1;
	  }
#if 0
	aim_conn_addhandler(ass, tstconn, AIM_CB_FAM_CTN, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
	aim_conn_addhandler(ass, tstconn, AIM_CB_FAM_GEN, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
#endif
	aim_conn_addhandler(ass, tstconn, 0x0001, 0x0003, at_serverready, 0);
	aim_auth_sendcookie(ass, tstconn, cookie);
//	fprintf(stderr, "\achatnav: connected\n");
      }
      break;
    case 0x000e: /* Chat */
	{
	char *roomname = NULL;
	struct aim_conn_t *tstconn = NULL;

	roomname = va_arg(ap, char *);

	tstconn = aim_newconn(ass, AIM_CONN_TYPE_CHAT, ip);
	if ( (tstconn==NULL) || (tstconn->status >= AIM_CONN_STATUS_RESOLVERR))
	  {
        log_debug(ZONE, "[AIM] Unable to connect to chat server");
        if(tstconn) aim_conn_kill(ass, &tstconn);
	    return 1;
	  }		
//	printf("AIM: chat: connected\n");

	/*
	 * We must do this to attach the stored name to the connection!
	 */
	aim_chat_attachname(tstconn, roomname);

	aim_conn_addhandler(ass, tstconn, 0x0001, 0x0003, at_serverready, 0);
	aim_auth_sendcookie(ass, tstconn, cookie);
    }
      break;
    default:
      log_warn("[AIM]", "uh oh... got redirect for unknown service 0x%04x!!\n", serviceid);
      /* dunno */
    }


	va_end(ap);
  return 1;
}


int at_parse_userinfo(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
  
#if 0
  char *sn = NULL;
  char *prof_encoding = NULL;
  char *prof = NULL;
  u_short warnlevel = 0x0000;
  u_short idletime = 0x0000;
  u_short class = 0x0000;
  u_long membersince = 0x00000000;
  u_long onlinesince = 0x00000000;

  va_list ap;
  va_start(ap, command);
  sn = va_arg(ap, char *);
  prof_encoding = va_arg(ap, char *);
  prof = va_arg(ap, char *);
  warnlevel = va_arg(ap, u_short);
  idletime = va_arg(ap, u_short);
  class = va_arg(ap, u_short);
  membersince = va_arg(ap, u_long);
  onlinesince = va_arg(ap, u_long);
  va_end(ap);
  
  printf("ac: userinfo: sn: %s\n", sn);
  printf("ac: userinfo: warnlevel: 0x%04x\n", warnlevel);
  printf("ac: userinfo: class: 0x%04x = ", class);

  /*
   *  00000000  (binary)
   *         1  Trial  
   *        2   Unknown
   *       3    AOL
   *      4     Unknown
   *     5      Free
   * 
   * ORed together.
   *
   */

  if (class & 0x0001)
    printf("TRIAL ");
  if (class & 0x0002)
    printf("UNKNOWN_BIT2 ");
  if (class & 0x0004)
    printf("AOL ");
  if (class & 0x0008)
    printf("UNKNOWN_BIT4 ");
  if (class & 0x0010)
    printf("FREE ");
  printf("\n");
  
  printf("ac: userinfo: membersince: %lu\n", membersince);
  printf("ac: userinfo: onlinesince: %lu\n", onlinesince);
  printf("ac: userinfo: idletime: 0x%04x\n", idletime);
  
  printf("ac: userinfo: profile_encoding: %s\n", prof_encoding ? prof_encoding : "[none]");
  printf("ac: userinfo: prof: %s\n", prof ? prof : "[none]");
#endif
  return 1;
}

int at_authsvrready(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
  //printf("at_authsvrready: called (contype: %d)\n", command->conn->type);
  //sleep(10);
  /* should just be able to tell it we're ready too... */
  aim_auth_clientready(ass, command->conn);

#if 0
  /*
   * This is where you'd really begin changing your password.
   *   However, this callback may get called for reasons other
   *   than you wanting to change your password.  You should 
   *   probably check that before actually doing it.
   */
  aim_auth_changepasswd(command->conn, "PWD1", "PWD2");
#endif

  return 1;
}

int pwdchngdone(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
  //printf("PASSWORD CHANGE SUCCESSFUL!!!\n");
  return 1;
}

int at_parse_oncoming(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{

	/* XXX mem usage? */
    xmlnode away;
    xmlnode status;
	xmlnode x;
	session s;
    jpacket jp;
    char *msg;
    struct aim_userinfo_s *userinfo;

    va_list ap;
    va_start(ap, command);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

	s = (session)ass->aux_data;

	x = xmlnode_new_tag("presence");
	xmlnode_put_attrib(x, "to", jid_full(s->j));
	xmlnode_put_attrib(x, "from", at_name);
	jp = jpacket_new(x);
	jid_set(jp->from, at_normalize(userinfo->sn), JID_USER);
	xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));
    if(userinfo->class&AIM_CLASS_AWAY)
    {
        away = xmlnode_insert_tag(x, "show");
        xmlnode_insert_cdata(away, "away", -1);
    }
    status = xmlnode_insert_tag(x, "status");
    msg = pmalloc(xmlnode_pool(x), 24);
    memset(msg, '\0', 24);
    sprintf(msg, "Idle %d Minutes", (userinfo->idletime));
    xmlnode_insert_cdata(status, msg, -1);
	ehandler_send(ehandler_at, jp->x, s->j->server);

	return 1;
}

int at_parse_offgoing(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{

	/* XXX mem usage? */
	xmlnode x;
	jpq q;
	jpacket jp;
	session s;

	s = (session)ass->aux_data;

	x = xmlnode_new_tag("presence");
	xmlnode_put_attrib(x, "to", jid_full(s->j));
	xmlnode_put_attrib(x, "from", at_name);
	xmlnode_put_attrib(x, "type","unavailable");
	jp = jpacket_new(x);
	jid_set(jp->from, at_normalize(&(command->data[11])), JID_USER);
	xmlnode_put_attrib(jp->x, "from", jid_full(jp->from));

	ehandler_send(ehandler_at, jp->x, s->j->server);

	//q = pmalloc(jp->p, sizeof(_jpq));
	//pth_msgport_put(s->mp_from, (pth_message_t *)q);
//	printf("\n%s has left\n", &(command->data[11]));

	return 1;
}

int at_parse_motd(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  char *msg;
  u_short id;
  va_list ap;
  
  va_start(ap, command);
  id = va_arg(ap, u_short);
  msg = va_arg(ap, char *);
  va_end(ap);

//  printf("AIM Trans: motd: %s\n", msg);

  return 1;
}

/* 
 * Handles callbacks for: AIM_CB_RATECHANGE, AIM_CB_USERERROR, 
 *   AIM_CB_MISSED_IM, and AIM_CB_MISSED_CALL.
 */
int at_parse_misses(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
  u_short family;
  u_short subtype;
  xmlnode x;
  xmlnode body;
	jpq q;
	jpacket jp;
	session s;
	int iserr=0;

	s = (session)ass->aux_data;

	x = xmlnode_new_tag("message");
	xmlnode_put_attrib(x, "to", jid_full(s->j));
	xmlnode_put_attrib(x, "from", at_name);
	xmlnode_put_attrib(x, "type","error");
	body = xmlnode_insert_tag(x, "error");

  family = (command->data[0] << 8) + command->data[1];
  subtype = (command->data[2] << 8) + command->data[3];
  
  switch (family)
    {
			// XXX I need to fix this to send error msgs
    case 0x0001:
      if (subtype == 0x000a) /* or AIM_CB_RATECHANGE */
      {
        xmlnode_put_attrib(body,"code","502"); /*remote server error*/
        xmlnode_insert_cdata(body, "You are sending messages too fast", 33);
	  iserr = 1;
      }
      break;
    case 0x0002:
      if (subtype == 0x0001) /* or AIM_CB_USERERROR */
	{
	  u_long snacid = 0x00000000;
	  
	  snacid = aimutil_get32(&command->data[6]);
        xmlnode_put_attrib(body,"code","502"); /*remote server error*/
	 xmlnode_insert_cdata(body, "Received unknown error in SNAC family 0x0002",
					 44);
	  iserr = 1;
	}
      break;
    case 0x0004:
      if (subtype == 0x0001) /* or AIM_CB_MISSED_IM */
      {
        xmlnode_put_attrib(body,"code","502"); /*remote server error*/
        xmlnode_insert_cdata(body, "The user you tried to message is not online", 43);
        iserr=1;
      }
      else if (subtype == 0x000a) /* or AIM_CB_MISSED_CALL */
      {
		printf("You missed some messages from %s because they were sent too fast\n", &(command->data[13]));
	  iserr = 1;
      }   
      break;
    default:
        xmlnode_put_attrib(body,"code","502"); /*remote server error*/
        xmlnode_insert_cdata(body,"An Unknown error has occured on the AIM server.",-1);
        break;
    }

  if(iserr)
  {
	jp = jpacket_new(x);
	ehandler_send(ehandler_at, jp->x, s->j->server);
	return 1;
	} else {
		xmlnode_free(body);
		xmlnode_free(x);
	}

  return 0;
}


int at_chat_join(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
    va_list ap;
    struct aim_userinfo_s *userinfo;
    int count = 0, i = 0;
  
    va_start(ap, command);
    count = va_arg(ap, int);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    log_debug(ZONE, "at: chat: %s:  New occupants have joined:\n", (char *)command->conn->priv);
    while (i < count)
        log_debug(ZONE, "at: chat: %s: \t%s\n", (char *)command->conn->priv, userinfo[i++].sn);

  return 1;
}

int at_chat_leave(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
    va_list ap;
    struct aim_userinfo_s *userinfo;
    int count = 0, i = 0;
  
    va_start(ap, command);
    count = va_arg(ap, int);
    userinfo = va_arg(ap, struct aim_userinfo_s *);
    va_end(ap);

    log_debug(ZONE, "at: chat: %s:  Some occupants have left:\n", (char *)command->conn->priv);
    while (i < count)
        log_debug(ZONE, "at: chat: %s: \t%s\n", (char *)command->conn->priv, userinfo[i++].sn);

  return 1;
}

int at_chat_infoupdate(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  va_list ap;
  struct aim_userinfo_s *userinfo;
  struct aim_chat_roominfo *roominfo;
  char *roomname;
  int usercount,i;
  char *roomdesc;

  va_start(ap, command);
  roominfo = va_arg(ap, struct aim_chat_roominfo *);
  roomname = va_arg(ap, char *);
  usercount= va_arg(ap, int);
  userinfo = va_arg(ap, struct aim_userinfo_s *);
  roomdesc = va_arg(ap, char *);
  va_end(ap);

//  printf("at: chat: %s:  info update:\n", (char *)command->conn->priv);
//  printf("at: chat: %s:  \tRoominfo: {%04x, %s, %04x}\n", 
//	(char *)command->conn->priv,
//	 roominfo->exchange,
//	 roominfo->name,
//	 roominfo->instance);
//  printf("at: chat: %s:  \tRoomname: %s\n", (char *)command->conn->priv, roomname);
//  printf("at: chat: %s:  \tRoomdesc: %s\n", (char *)command->conn->priv, roomdesc);
//  printf("at: chat: %s:  \tOccupants: (%d)\n", (char *)command->conn->priv, usercount);
  
  //i = 0;
//  while (i < usercount)
//    printf("at: chat: %s:  \t\t%s\n", (char *)command->conn->priv, userinfo[i++].sn);

  return 1;
}

int at_chat_incomingmsg(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  va_list ap;
  struct aim_userinfo_s *userinfo;
  char *msg;
  char tmpbuf[1152];
 
  va_start(ap, command);
  userinfo = va_arg(ap, struct aim_userinfo_s *);	
  msg = va_arg(ap, char *);
  va_end(ap);

    log_debug(ZONE,"at: chat: %s: incoming msg from %s: %s\n", (char *)command->conn->priv, userinfo->sn, msg);

  /*
   * Do an echo for testing purposes.  But not for ourselves ("oops!")
   */
/*  if (strcmp(userinfo->sn, sess->logininfo.screen_name) != 0)
    {
      sprintf(tmpbuf, "(%s said \"%s\")", userinfo->sn, msg);
      aim_chat_send_im(sess, command->conn, tmpbuf);
    }*/

  return 1;
}

int at_chatnav_info(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  u_short type;
  va_list ap;

  ap = va_start(ap, command);
  type = va_arg(ap, u_short);

  switch(type)
    {
    case 0x0002:
      {
	int maxrooms;
	struct aim_chat_exchangeinfo *exchanges;
	int exchangecount,i = 0;
	
	maxrooms = va_arg(ap, u_char);
	exchangecount = va_arg(ap, int);
	exchanges = va_arg(ap, struct aim_chat_exchangeinfo *);
	va_end(ap);

	printf("at: chat info: Chat Rights:\n");
	printf("at: chat info: \tMax Concurrent Rooms: %d\n", maxrooms);
	
	printf("at: chat info: \tExchange List: (%d total)\n", exchangecount);
	while (i < exchangecount)
	  {
	    printf("at: chat info: \t\t%x: %s (%s/%s)\n", 
		   exchanges[i].number,	
		   exchanges[i].name,
		   exchanges[i].charset1,
		   exchanges[i].lang1);
	    i++;
	  }
	
      }
      break;
    default:
      va_end(ap);
      printf("at: chatnav info: unknown type (%04x)\n", type);
    }
  return 1;
}

int at_parse_incoming_im(struct aim_session_t *ass, struct command_rx_struct *command, ...)
{
	int channel;
	va_list ap;

	va_start(ap, command);
	channel = va_arg(ap, int);

	/*
	* Channel 1: Standard Message
	*/
	if (channel == 1)
	{
    	struct aim_userinfo_s *userinfo;
	    char *msg = NULL;
    	u_int icbmflags = 0;
    	u_short flag1, flag2;
		jpacket jp;
		xmlnode par;
		xmlnode node;
		session s;
		jid from;
  
	    userinfo = va_arg(ap, struct aim_userinfo_s *);
    	msg = va_arg(ap, char *);
	    icbmflags = va_arg(ap, u_int);
    	flag1 = va_arg(ap, u_short);
	    flag2 = va_arg(ap, u_short);
    	va_end(ap);

		s = (session)ass->aux_data;


        /* XXX: This is butt ugly until the rewrite */
		par = xmlnode_new_tag("message");
		xmlnode_put_attrib(par, "type", "chat");
		xmlnode_put_attrib(par, "to", jid_full(s->j));
		xmlnode_put_attrib(par, "from", at_name);
		jp = jpacket_new(par);
		node = xmlnode_insert_tag(jp->x, "body");
        msg = (char *)strip_html(msg, jp->p);
		msg = (char *)str_to_UTF8(jp->p, (unsigned char *)msg);
		xmlnode_insert_cdata(node, j_strdup(msg), strlen(msg));


		jid_set(jp->from, at_normalize((char *)userinfo->sn), JID_USER);
                xmlnode_put_attrib(jp->x,"from",jid_full(jp->from));

		log_debug(ZONE, "[AIM] Sending: %s\n", xmlnode2str(jp->x));
		ehandler_send(ehandler_at, jp->x, s->j->server);
		
        /* XXX I'm dumb, left for stupid reflection
         * the pool frees this
         *
         * free(msg);
         *
         */

	}
	else if(channel == 2)
	{
		/* XXX decide how to handle groupchat requests */
		/* XXX Actualy parse and use this */
	}

//  printf("at: icbm: done with ICBM handling\n");

  return 1;
}

int at_parse_authresp(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  struct aim_conn_t *bosconn = NULL;
  

//  printf("Screen name: %s\n", sess->logininfo.screen_name);

  /*
   * Check for error.
   */
  if (sess->logininfo.errorcode)
    {
		jpq q;
		jpacket jp;
		session s;
		xmlnode x;
		xmlnode body;

		s = (session)sess->aux_data;

		x = xmlnode_new_tag("message");
		xmlnode_put_attrib(x, "to", jid_full(s->j));
		xmlnode_put_attrib(x, "from", at_name);
		xmlnode_put_attrib(x, "type","error");
		body = xmlnode_insert_tag(x, "error");
		// Make this more desscriptive
		xmlnode_insert_cdata(body, "There was an error logging in to AIM", 36);
		jp = jpacket_new(x);
		ehandler_send(ehandler_at, jp->x, s->j->server);

      //printf("Login Error Code 0x%14x\n", sess->logininfo.errorcode);
        log_debug(ZONE,"[AIM] Error URL: %s\n", sess->logininfo.errorurl);
      aim_conn_kill(sess, &command->conn);
	  return 0;
      /* exit(0); */ /* XXX: should return in order to let the above things get free()'d. */
    }

//  printf("Reg status: %2d\n", sess->logininfo.regstatus);
//  printf("Email: %s\n", sess->logininfo.email);
//  printf("BOS IP: %s\n", sess->logininfo.BOSIP);

//  printf("Closing auth connection...\n");
  aim_conn_kill(sess, &command->conn);
  bosconn = aim_newconn(sess, AIM_CONN_TYPE_BOS, sess->logininfo.BOSIP);
  if (bosconn == NULL)
    {
      log_error(ZONE, "[AIM] could not connect to BOS: internal error\n");
    }
  else if (bosconn->status != 0)
    {
      log_error(ZONE, "[AIM] could not connect to BOS\n");
        aim_conn_kill(sess, &command->conn);
    }
  else
    {
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_ACK, AIM_CB_ACK_ACK, NULL, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_SERVERREADY, at_serverready, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATEINFO, NULL, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_REDIRECT, at_handleredirect, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_STS, AIM_CB_STS_SETREPORTINTERVAL, NULL, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_BUD, AIM_CB_BUD_ONCOMING, at_parse_oncoming, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_BUD, AIM_CB_BUD_OFFGOING, at_parse_offgoing, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_INCOMING, at_parse_incoming_im, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_LOC, AIM_CB_LOC_ERROR, at_parse_misses, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_MISSEDCALL, at_parse_misses, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_RATECHANGE, at_parse_misses, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_MSG, AIM_CB_MSG_ERROR, at_parse_misses, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_LOC, AIM_CB_LOC_USERINFO, at_parse_userinfo, 0);

      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_CTN, AIM_CB_CTN_DEFAULT, aim_parse_unknown, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_DEFAULT, aim_parse_unknown, 0);
      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_GEN, AIM_CB_GEN_MOTD, at_parse_motd, 0);

      aim_conn_addhandler(sess, bosconn, AIM_CB_FAM_SPECIAL, AIM_CB_SPECIAL_CONNERR, at_parse_connerr, 0);

      aim_auth_sendcookie(sess, bosconn, sess->logininfo.cookie);
    }
  return 1;
}

struct aim_conn_t *_aim_select(struct aim_session_t *ass, pth_event_t ev, int *status)
{
  pth_event_t tst;
  fd_set fds;
  fd_set errfds;
  int i;
  int maxfd = 0;
  struct aim_conn_t *cur;

  /* Reset status */
  *status = 0;

  if (aim_countconn(ass) <= 0)
    return 0;

  /* 
   * If we have data waiting to be sent, return immediatly
   */
    if (ass->queue_outgoing)
    {
        *status = 1;
        return (struct aim_conn_t *)1;
    }

    FD_ZERO(&fds);
  
    for (cur = ass->connlist; cur; cur = cur->next) 
    {
        FD_SET(cur->fd, &fds);
        if (cur->fd > maxfd)
            maxfd = cur->fd;
    }
  
    i = pth_select_ev(maxfd+1, &fds, NULL, NULL, NULL, ev);
	if (i>=1)
	{
        for (cur = ass->connlist; cur; cur = cur->next) 
        {
            if (FD_ISSET(cur->fd, &fds)) 
            {
                *status = 2;
                return cur; 
            }
        }
	}
    if(pth_event_occurred(ev))
    {
        /* Possible msgport event (should be, no timeout) */
        *status = 3;
        return (struct aim_conn_t *)1;
    }

    *status = i; /* 0 or -1 */
	return (struct aim_conn_t *)1; /* NO REACH */
}


int at_parse_connerr(struct aim_session_t *sess, struct command_rx_struct *command, ...)
{
  va_list ap;
  unsigned short code;
  char *msg = NULL;

  ap = va_start(ap, command);
  code = va_arg(ap, unsigned short);
  msg = va_arg(ap, char *);
  va_end(ap);

  log_debug(ZONE,"AT: faimtest: connerr: Code 0x%04x: %s\n", code, msg);

  return 1;
}

