#include <aimtrans.h>

void at_parse_packet(xmlnode x)
{
    jpacket jp;
    session s;
    xmlnode dup;

    log_debug(ZONE, "[AIM] Parsing Packet\n");
        
    jp = jpacket_new(x);
    s = at_session_find_by_jid(jp->from);
    if(s != NULL)
    {
        log_debug(ZONE, "[AIM] Sent to session");
        at_psend(s->mp_to, jp);
        return;
    }
    switch(jp->type)
    {
        case JPACKET_IQ:
            if(j_strcmp(xmlnode_get_attrib(xmlnode_get_tag(jp->x,"query"),"xmlns"),"jabber:iq:register") == 0)
            {
                /* Register the user */
                log_debug(ZONE, "[AIM] Sent to the register");
				at_auth_register(jp);
                break;
            }
            if(jp->subtype == JPACKET__GET)
            {
                if( j_strcmp( jp->iqns, "jabber:iq:version" ) == 0 )
                {
                    xmlnode iq, iqname;
                    log_debug(ZONE, "[AIM] Responding to version query\n");
                    iq = xmlnode_dup( jp->x );
                    xmlnode_put_attrib( iq, "to", jid_full(jp->from) );
                    xmlnode_put_attrib( iq, "from", jid_full(jp->to) );
                    xmlnode_put_attrib( iq, "type", "result" );
                    iqname = xmlnode_insert_tag( iq, "name" );
                    xmlnode_insert_cdata( iqname, "Jabber AIM Transport", -1);
                    ehandler_send( ehandler_at, iq, jp->from->server );
                }
                else if( j_strcmp( jp->iqns, "jabber:iq:time" ) == 0 )
                {
                    xmlnode iq;
                    log_debug(ZONE, "[AIM] Responding to time query\n");
                    iq = jutil_iqresult( jp->x );
                    xmlnode_put_attrib(iq,"to",jid_full(jp->from));
                    xmlnode_put_attrib(iq,"from",jid_full(jp->to));
                    xmlnode_insert_cdata( xmlnode_insert_tag( iq, "utc"),jutil_timestamp(),-1);
                    xmlnode_insert_cdata(xmlnode_insert_tag(iq,"tz"),"gmt",-1);
                    ehandler_send(ehandler_at,iq,jp->from->server);
                }
            } 
            break;
        case JPACKET_S10N:
            /* Handle a S10N packet based on their subtype */
            log_debug(ZONE, "[AIM] We got a s10n packet");
            switch(jpacket_subtype(jp))
            {
            case JPACKET__SUBSCRIBED:
                /* Do I need to do anything? */
                break;
            case JPACKET__UNSUBSCRIBED:
                /* Do I need to do anything? */
                break;
            case JPACKET__SUBSCRIBE:
                log_debug(ZONE, "[AIM] new subscription\n");
                if(jp->to->resource == NULL)
                    at_auth_add_buddy(jp);
                break;
            case JPACKET__UNSUBSCRIBE:
                /* Do more with this 
                * (Actually take the user off the aim roster)
               * should we check for a subscription as well? 
                */
                dup = xmlnode_dup(jp->x);

                s=at_session_find_by_jid(jp->from);
                if(s==NULL)
                {
                    jutil_error(dup, TERROR_REGISTER);
                    ehandler_send(ehandler_at, dup, jp->from->server);
                    break;
                }
                at_session_buddy_unlink(s, jp->to->user);
                aim_remove_buddy(&s->ass, aim_getconn_type(&s->ass,
                                AIM_CONN_TYPE_BOS), jp->to->user);
                log_debug(ZONE, "[AIM] Unsubscribing\n");
                xmlnode_put_attrib(dup, "type", "unsubscribed");
                xmlnode_put_attrib(dup, "to", jid_full(jp->from));
                xmlnode_hide_attrib(dup, "from");
                ehandler_send(ehandler_at, dup, jp->from->server);
                break;
			}
        case JPACKET_PRESENCE:
            if(jp->subtype == JPACKET__AVAILABLE)
            {
                if(jp->to->resource == NULL)
                {
                   at_auth_add_buddy(jp);
                   return;
                } else {
                    log_debug(ZONE, "[AIM] Logging in user");
                    at_auth_user(xmlnode_dup(jp->x));
                }
            }
            break;
        default:
            log_debug(ZONE, "Could not handle %s", xmlnode2str(jp->x));
            break;
	}

    xmlnode_free(jp->x);

	return;
}
